const fs = require('fs')
const path = require('path')
const Config = require('webpack-chain')
const MiniCssExtractPlugin = require('mini-css-extract-plugin')

// 项目下是否有配置表，先加载，方便修改修改paths配置
let projectConfig = null
const projectConfigPath = path.resolve(process.cwd(), 'webpack.chain.js')
if (fs.existsSync(projectConfigPath)) {
  projectConfig = require(projectConfigPath)
}

const paths = require('./paths')

const genAssetPath = (dir) => {
  return process.env.APP_MODE === 'production'
    ? `${paths.assetsSubPath}${dir}[name].[contenthash:8].[ext]`
    : '[name].[ext]'
}

const genUrlLoaderOptions = (dir) => {
  return {
    limit: 1024 * 8,
    name: genAssetPath(dir)
  }
}

/**
 * 创建webpack配置
 * @return { config, paths }
 */
module.exports = () => {
  const config = new Config()
  const MODE_PROD = process.env.APP_MODE === 'production'

  config
    .mode(process.env.APP_MODE)
    .entry('main')
    .add(paths.appEntry)
    .end()
    .output.path(paths.appBuild)
    .filename(
      MODE_PROD
        ? `${paths.assetsSubPath}${paths.assetsJsPath}[name].[contenthash:8].js`
        : '[name].js'
    )
    .publicPath(paths.publicPath)
    .end()
    .resolve.alias.set('@', paths.appSrc)
    .end()
    .extensions.merge(['.ts', '.tsx', '.js'])

  /**
   * webpack module
   */
  config.module
    .rule('babel')
    .test(/\.(j|t)sx?$/)
    .include.add(paths.appSrc)
    .add('/node_modules/@hackkun/react-router-kit')
    .end()
    .use('babel-loader')
    .loader('babel-loader')
    .options({
      presets: [
        [
          '@babel/preset-env',
          {
            modules: false
          }
        ],
        '@babel/preset-react',
        '@babel/preset-typescript'
      ],
      plugins: [
        [
          '@babel/plugin-transform-runtime',
          {
            corejs: {
              version: 3,
              proposals: true
            },
            useESModules: true
          }
        ],
        '@babel/plugin-syntax-dynamic-import',
        ['@babel/plugin-proposal-decorators', { legacy: true }],
        ['@babel/plugin-proposal-class-properties', { loose: true }],
        [
          'import',
          {
            libraryName: 'antd',
            libraryDirectory: 'es',
            style: true
          }
        ]
      ]
    })
    .end()
    .end()
    .rule('less')
    .test(/\.(c|le)ss$/)
    .use('style-loader')
    .loader(MODE_PROD ? MiniCssExtractPlugin.loader : 'style-loader')
    .end()
    .use('css-loader')
    .loader('css-loader')
    .end()
    .use('postcss-loader')
    .loader('postcss-loader')
    .options({
      plugins: [require('autoprefixer')({})]
    })
    .end()
    .use('less-loader')
    .loader('less-loader')
    .options({
      lessOptions: {
        javascriptEnabled: true,
        modifyVars: {}
      }
    })
    .end()
    .end()
    .rule('img')
    .test(/\.(png|jpe?g|gif|webp)(\?.*)?$/)
    .use('imgLoader')
    .loader('url-loader')
    .options(genUrlLoaderOptions(paths.assetsImgPath))
    .end()
    .end()
    .rule('svg')
    .test(/\.(svg)(\?.*)?$/)
    .use('svgLoader')
    .loader('file-loader')
    .options({
      name: genAssetPath(paths.assetsImgPath)
    })
    .end()
    .end()
    .rule('font')
    .test(/\.(woff2?|eot|ttf|otf)(\?.*)?$/i)
    .use('fontLoader')
    .loader('url-loader')
    .options(genUrlLoaderOptions(paths.assetsFontPath))
    .end()
    .end()
    .rule('media')
    .test(/\.(mp4|webm|ogg|mp3|wav|flac|aac)(\?.*)?$/)
    .use('mediaLoader')
    .loader('url-loader')
    .options(genUrlLoaderOptions(paths.assetsMediaDir))

  /**
   * webpack plugins
   */
  config
    .plugin('env')
    .use(require('dotenv-webpack'), [
      {
        path: `.env.${process.env.APP_ENV}`
      }
    ])
    .end()
    .plugin('html')
    .use(require('html-webpack-plugin'), [
      {
        template: paths.appTemplate,
        filename: 'index.html'
      }
    ])
    .end()
    .plugin('copy')
    .use(require('copy-webpack-plugin'), [
      {
        patterns: [
          {
            from: paths.appStatic,
            to: paths.appBuild,
            globOptions: {
              ignore: ['index.html']
            }
          }
        ]
      }
    ])

  /**
   * development 配置
   */
  config.when(process.env.APP_MODE === 'development', (config) => {
    config.plugin('error').use(require('friendly-errors-webpack-plugin'))
    config
      .devtool('cheap-module-eval-source-map')
      .devServer.port(4000)
      .hot(true)
      .quiet(true)
      .host('0.0.0.0')
      .historyApiFallback(true)
      .clientLogLevel('none')
      .contentBase(paths.appBuild)
      .publicPath(paths.publicPath)
  })

  /**
   * production 配置
   */
  config.when(MODE_PROD, (config) => {
    config
      .plugin('progress')
      .use(require('webpackbar'), [
        {
          name: '开始构建项目'
        }
      ])
      .end()
      .plugin('css-extract')
      .use(MiniCssExtractPlugin, [
        {
          filename: `${paths.assetsSubPath}${paths.assetsCssPath}[name].[contenthash:8].css`
        }
      ])

    /**
     * webpack压缩
     */
    config.performance
      .hints(false)
      .end()
      .optimization.runtimeChunk({
        name: 'runtime'
      })
      .minimizer('mini-css')
      .use(require('optimize-css-assets-webpack-plugin'))
      .end()
      .splitChunks({
        chunks: 'all',
        cacheGroups: {
          vendors: {
            name: 'vendor',
            test: /[\\/]node_modules[\\/]/,
            priority: -10,
            chunks: 'initial'
          },
          common: {
            name: 'common',
            minChunks: 2,
            priority: -20,
            chunks: 'initial'
          }
        }
      })
  })

  if (typeof projectConfig === 'function') {
    projectConfig(config, process.env.APP_MODE)
  }

  return { config, paths }
}
